"""Cog wrapper for command caching."""
from datetime import datetime, timedelta, timezone

from dis_snek import InteractionContext, Scale, Snake
from dis_snek.client.utils.misc_utils import find
from dis_snek.models.discord.embed import EmbedField
from dis_snek.models.snek.tasks.task import Task
from dis_snek.models.snek.tasks.triggers import IntervalTrigger
from jarvis_core.db import q
from jarvis_core.db.models import (
    Action,
    Ban,
    Kick,
    Modlog,
    Mute,
    Note,
    Setting,
    Warning,
)

from jarvis.utils import build_embed

MODLOG_LOOKUP = {"Ban": Ban, "Kick": Kick, "Mute": Mute, "Warning": Warning}
IGNORE_COMMANDS = {"Ban": ["bans"], "Kick": [], "Mute": ["unmute"], "Warning": ["warnings"]}


class CacheCog(Scale):
    """Cog wrapper for command caching."""

    def __init__(self, bot: Snake):
        self.bot = bot
        self.cache = {}
        self._expire_interaction.start()

    def check_cache(self, ctx: InteractionContext, **kwargs: dict) -> dict:
        """Check the cache."""
        if not kwargs:
            kwargs = {}
        return find(
            lambda x: x["command"] == ctx.subcommand_name  # noqa: W503
            and x["user"] == ctx.author.id  # noqa: W503
            and x["guild"] == ctx.guild.id  # noqa: W503
            and all(x[k] == v for k, v in kwargs.items()),  # noqa: W503
            self.cache.values(),
        )

    @Task.create(IntervalTrigger(minutes=1))
    async def _expire_interaction(self) -> None:
        keys = list(self.cache.keys())
        for key in keys:
            if self.cache[key]["timeout"] <= datetime.now(tz=timezone.utc) + timedelta(minutes=1):
                del self.cache[key]


class ModcaseCog(Scale):
    """Cog wrapper for moderation case logging."""

    def __init__(self, bot: Snake):
        self.bot = bot
        self.add_scale_postrun(self.log)

    async def log(self, ctx: InteractionContext, *args: list, **kwargs: dict) -> None:
        """
        Log a moderation activity in a moderation case.

        Args:
            ctx: Command context
        """
        name = self.__name__.replace("Cog", "")

        if name in MODLOG_LOOKUP and ctx.command not in IGNORE_COMMANDS[name]:
            user = kwargs.pop("user", None)
            if not user and not ctx.target_id:
                self.logger.warn(f"Admin action {name} missing user, exiting")
                return
            elif ctx.target_id:
                user = ctx.target
            coll = MODLOG_LOOKUP.get(name, None)
            if not coll:
                self.logger.warn(f"Unsupported action {name}, exiting")
                return

            action = await coll.find_one(q(user=user.id, guild=ctx.guild_id, active=True))
            if not action:
                self.logger.warn(f"Missing action {name}, exiting")
                return

            action = Action(action_type=name.lower(), parent=action.id)
            note = Note(admin=self.bot.user.id, content="Moderation case opened automatically")
            await Modlog(user=user.id, admin=ctx.author.id, actions=[action], notes=[note]).commit()
            notify = await Setting.find_one(q(guild=ctx.guild.id, setting="notify", value=True))
            if notify and name not in ["Kick", "Ban"]:  # Ignore Kick and Ban, as these are unique
                fields = [
                    EmbedField(name="Action Type", value=name, inline=False),
                    EmbedField(
                        name="Reason", value=kwargs.get("reason", None) or "N/A", inline=False
                    ),
                ]
                embed = build_embed(
                    title="Admin action taken",
                    description=f"Admin action has been taken against you in {ctx.guild.name}",
                    fields=fields,
                )
                guild_url = f"https://discord.com/channels/{ctx.guild.id}"
                embed.set_author(name=ctx.guild.name, icon_url=ctx.guild.icon.url, url=guild_url)
                embed.set_thumbnail(url=ctx.guild.icon.url)
                await user.send(embed=embed)
